#include <windows.h>

#include <iostream>
using namespace std;

struct DebugStatusRegister
{
	ULONG L0:1;
	ULONG G0:1;
	ULONG L1:1;
	ULONG G1:1;
	ULONG L2:1;
	ULONG G2:1;
	ULONG L3:1;
	ULONG G3:1;
	ULONG LE:1;
	ULONG GE:1;
	ULONG:3;
	ULONG GD:1;
	ULONG:2;
	ULONG RW0:2;
	ULONG LEN0:2;
	ULONG RW1:2;
	ULONG LEN1:2;
	ULONG RW2:2;
	ULONG LEN2:2;
	ULONG RW3:2;
	ULONG LEN3:2;
};

void dump()
{
	cout<<"Dumping..."<<endl;

	cout<<"  Creating kmd.exe process..."<<endl;

	STARTUPINFO si;
	ZeroMemory(&si,sizeof si);
	si.cb=sizeof si;

	PROCESS_INFORMATION pi;

	if (!CreateProcess("kmd.exe",NULL,NULL,NULL,false,DEBUG_PROCESS,NULL,NULL,&si,&pi))
	{
		cout<<"    ERROR: Couldn't start kmd.exe."<<endl;
		cout<<"    Run this program from a directory containing kmd.exe."<<endl;
		exit(-1);
	}

	CloseHandle(pi.hProcess);
	CloseHandle(pi.hThread);

	DEBUG_EVENT de;
	HANDLE hThread;
	HANDLE hProcess;

	do
	{
		WaitForDebugEvent(&de,INFINITE);

		DWORD dwContinueStatus=DBG_EXCEPTION_NOT_HANDLED;

		switch (de.dwDebugEventCode)
		{
		case CREATE_PROCESS_DEBUG_EVENT:
			hThread=de.u.CreateProcessInfo.hThread;
			hProcess=de.u.CreateProcessInfo.hProcess;
			break;
		case EXIT_PROCESS_DEBUG_EVENT:
			cout<<"    ERROR: Process exited prematurely."<<endl;
			exit(-1);
		case EXCEPTION_DEBUG_EVENT:
			if  (de.u.Exception.ExceptionRecord.ExceptionCode==EXCEPTION_BREAKPOINT)
			{
				dwContinueStatus=DBG_CONTINUE;
				
				//I won't explain why this works, but it does.
				cout<<"    Killing anti-debugging code..."<<endl;
				WriteProcessMemory(hProcess,(void*)0x0043C053,"\0",1,NULL);
			}

			{
				//place/replace the breakpoint
				cout<<"    Setting breakpoint..."<<endl;
				CONTEXT c;
				c.ContextFlags=CONTEXT_DEBUG_REGISTERS;
				GetThreadContext(hThread,&c);
				c.Dr0=0x00430C4C;
				DebugStatusRegister* pDC=(DebugStatusRegister*)&c.Dr7;
				pDC->L0=1;
				pDC->LEN0=0; //length == 1 byte
				pDC->RW0=1; //break on writes
				SetThreadContext(hThread,&c);
			}
		}

		ContinueDebugEvent(de.dwProcessId,de.dwThreadId,dwContinueStatus);
	}
	while (de.dwDebugEventCode!=EXCEPTION_DEBUG_EVENT || de.u.Exception.ExceptionRecord.ExceptionCode!=EXCEPTION_SINGLE_STEP);

	//we've hit the break-point, so write the image
	cout<<"  Writing image to disk from memory..."<<endl;

	const int nImageSize=245760;

	void* buf=malloc(nImageSize);
	ReadProcessMemory(hProcess,(void*)0x00400000,buf,nImageSize,NULL);

	HANDLE hFile=CreateFile("kmd.unpacked.exe",GENERIC_WRITE,0,NULL,CREATE_ALWAYS,FILE_ATTRIBUTE_NORMAL,NULL);
	DWORD dwBytesWritten;
	WriteFile(hFile,buf,nImageSize,&dwBytesWritten,NULL);
	CloseHandle(hFile);

	free(buf);
}

void decrypt(unsigned char* pszName)
{
	DWORD key=0x7AF8D724;

	while (true)
	{
		*pszName^=(unsigned char)key;
		*pszName+=(unsigned char)(key>>8);
		*pszName-=(unsigned char)key;
		if (!*pszName)
			break;

		__asm rol key,0x86;

		pszName++;
	}
}

void patch()
{
	cout<<"Patching..."<<endl;

	HANDLE hFile=CreateFile("kmd.unpacked.exe",GENERIC_READ|GENERIC_WRITE,FILE_SHARE_READ,NULL,OPEN_EXISTING,FILE_ATTRIBUTE_NORMAL,NULL);
	HANDLE hMap=CreateFileMapping(hFile,NULL,PAGE_READWRITE,0,0,NULL);
	char* pFile=(char*)MapViewOfFile(hMap,FILE_MAP_WRITE,0,0,0);

	cout<<"  Decrypting import names..."<<endl;
	DWORD* pdwIAT=(DWORD*)(pFile+0x25000);
	for (int i=0;i<0x2C0/sizeof DWORD;i++)
	{
		if (pdwIAT[i]<0x8000000 && pdwIAT[i]!=0)
		{
			char* pszName=pFile+pdwIAT[i];
			decrypt((unsigned char*)pszName);
		}
	}

	cout<<"  Decrypting import lookup table..."<<endl;
	//isn't this easy :-)
	for (i=0x28F3C;i<0x291FC;i++)
	{
		pFile[i]=~pFile[i];
	}

	cout<<"  Offsetting import name table pointers..."<<endl;
	pdwIAT=(DWORD*)(pFile+0x25000);
	for (i=0;i<0x2C0/sizeof DWORD;i++)
	{
		if (pdwIAT[i]<0x8000000 && pdwIAT[i]!=0)
			pdwIAT[i]=pdwIAT[i]-2;
	}
	pdwIAT=(DWORD*)(pFile+0x28F3C);
	for (i=0;i<0x2C0/sizeof DWORD;i++)
	{
		if (pdwIAT[i]<0x8000000 && pdwIAT[i]!=0)
			pdwIAT[i]=pdwIAT[i]-2;
	}

	//PATCH HEADER
	cout<<"  Fixing headers..."<<endl;

	cout<<"    Fixing number of sections..."<<endl;
	*(short*)(pFile+0xEE)=4;

	cout<<"    Fixing entry point RVA..."<<endl;
	*(int*)(pFile+0x110)=0x0001A9C0;

	cout<<"    Fixing import table RVA & size..."<<endl;
	*(int*)(pFile+0x168)=0x00028E88;
	*(int*)(pFile+0x16C)=0xD8D;

	cout<<"    Fixing image size..."<<endl;
	*(int*)(pFile+0x138)=0x0003C000;

	cout<<"    Fixing section 1 header..."<<endl;
	strcpy(pFile+0x1E0,".text");
	*(unsigned int*)(pFile+0x1E0+0x10)=*(unsigned int*)(pFile+0x1E0+0x8);
	*(unsigned int*)(pFile+0x1E0+0x14)=*(unsigned int*)(pFile+0x1E0+0xC);
	*(unsigned int*)(pFile+0x1E0+0x24)=0x60000020;

	cout<<"    Fixing section 2 header..."<<endl;
	strcpy(pFile+0x208,".rdata");
	*(unsigned int*)(pFile+0x208+0x10)=*(unsigned int*)(pFile+0x208+0x8);
	*(unsigned int*)(pFile+0x208+0x14)=*(unsigned int*)(pFile+0x208+0xC);
	*(unsigned int*)(pFile+0x208+0x24)=0x40000040;

	cout<<"    Fixing section 3 header..."<<endl;
	strcpy(pFile+0x230,".data");
	*(unsigned int*)(pFile+0x230+0x10)=*(unsigned int*)(pFile+0x230+0x8);
	*(unsigned int*)(pFile+0x230+0x14)=*(unsigned int*)(pFile+0x230+0xC);
	*(unsigned int*)(pFile+0x230+0x24)=0xC0000040;

	cout<<"    Fixing section 4 header..."<<endl;
	strcpy(pFile+0x258,".rsrc");
	*(unsigned int*)(pFile+0x258+0x10)=*(unsigned int*)(pFile+0x258+0x8);
	*(unsigned int*)(pFile+0x258+0x14)=*(unsigned int*)(pFile+0x258+0xC);
	*(unsigned int*)(pFile+0x258+0x24)=0x40000040;

	cout<<"    Removing section 5 header..."<<endl;
	memset(pFile+0x280,0,40);

	UnmapViewOfFile(pFile);
	CloseHandle(hMap);
	CloseHandle(hFile);
}

int main(int argc, char* argv[])
{
	dump();
	patch();
	cout<<"Done!"<<endl;

	return 0;
}
